import logging

logger = logging.getLogger("vlad_logger")
logger.setLevel(logging.INFO)
sh = logging.StreamHandler()
sh.setLevel(logging.INFO)
sh.setFormatter(logging.Formatter("%(message)s"))
logger.addHandler(sh)
logger.propagate = False



import io

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from vladiate.exceptions import MissingExtraException


class VladInput(object):
    """A generic input class"""

    def __init__(self):
        raise NotImplementedError

    def open(self):
        raise NotImplementedError

    def __repr__(self):
        raise NotImplementedError


class LocalFile(VladInput):
    """Read from a local file path"""

    def __init__(self, filename):
        self.filename = filename

    def open(self):
        with open(self.filename, "r") as f:
            return f.readlines()

    def __repr__(self):
        return "{}('{}')".format(self.__class__.__name__, self.filename)


class S3File(VladInput):
    """Read from a file in S3"""

    def __init__(self, path=None, bucket=None, key=None):
        try:
            import boto  # noqa

            self.boto = boto
        except ImportError:
            # 2.7 workaround, should just be `raise Exception() from None`
            exc = MissingExtraException()
            exc.__context__ = None
            raise exc

        if path and not any((bucket, key)):
            self.path = path
            parse_result = urlparse(path)
            self.bucket = parse_result.netloc
            self.key = parse_result.path
        elif all((bucket, key)):
            self.bucket = bucket
            self.key = key
            self.path = "s3://{}{}"
        else:
            raise ValueError(
                "Either 'path' argument or 'bucket' and 'key' argument must " "be set."
            )

    def open(self):
        s3 = self.boto.connect_s3()
        bucket = s3.get_bucket(self.bucket)
        key = bucket.new_key(self.key)
        contents = key.get_contents_as_string()
        ret = io.BytesIO(bytes(contents))
        return ret

    def __repr__(self):
        return "{}('{}')".format(self.__class__.__name__, self.path)


class String(VladInput):
    """Read a file from a string"""

    def __init__(self, string_input=None, string_io=None):
        self.string_io = string_io if string_io else StringIO(string_input)

    def open(self):
        return self.string_io

    def __repr__(self):
        return "{}('{}')".format(self.__class__.__name__, "...")



import pytest
from pretend import stub, call, call_recorder

from vladiate.exceptions import BadValidatorException, ValidationException
from vladiate.validators import (
    CastValidator,
    EmptyValidator,
    FloatValidator,
    Ignore,
    IntValidator,
    NotEmptyValidator,
    RangeValidator,
    RegexValidator,
    RowValidator,
    RowLengthValidator,
    SetValidator,
    UniqueValidator,
    Validator,
    _stringify_set,
)


class FakeRow(object):
    """A row with side effects"""

    def __init__(self, fields):
        self.fields = fields

    def __getitem__(self, key):
        return self.fields[key].pop(0)

    def keys(self):
        return self.fields.keys()


def test_cast_validator():
    def _raise_valueerror(x):
        raise ValueError

    validator = CastValidator()
    validator.cast = call_recorder(_raise_valueerror)

    assert validator.bad == set()

    with pytest.raises(ValidationException):
        validator.validate("field", "something")

    assert validator.bad == {"field"}

    assert validator.cast.calls == [call("field")]


@pytest.mark.parametrize("field", [("42"), ("42.0"), ("-42.0")])
def test_float_validator_works(field):
    FloatValidator().validate(field)


@pytest.mark.parametrize("field", [("foo"), (" "), ("7..0"), ("4,200")])
def test_float_validator_fails(field):
    with pytest.raises(ValidationException):
        FloatValidator().validate(field)


@pytest.mark.parametrize("field", [("42")])
def test_int_validator_works(field):
    IntValidator().validate(field)


@pytest.mark.parametrize(
    "field", [("foo"), (" "), ("7..0"), ("4,200"), ("42.0"), ("-42.0")]
)
def test_int_validator_fails(field):
    with pytest.raises(ValidationException):
        IntValidator().validate(field)


@pytest.mark.parametrize(
    "field_set, field", [(["foo"], "foo"), (["foo", "bar"], "foo")]
)
def test_set_validator_works(field_set, field):
    validator = SetValidator(field_set)
    validator.validate(field)
    assert field in validator.valid_set


@pytest.mark.parametrize(
    "field_set, field",
    [
        (["foo"], "Foo"),
        (["Bar"], "bar"),
        (["foo"], "foo"),
        (["Foo", "Bar"], "foo"),
        (["Foo", "Bar"], "bar"),
    ],
)
def test_set_validator_supports_ignore_case(field_set, field):
    validator = SetValidator(field_set, ignore_case=True)
    field_to_check = field.lower() if validator.ignore_case else field
    validator.validate(field_to_check)
    assert field_to_check in validator.set_to_check


@pytest.mark.parametrize(
    "field_set, field", [([], "bar"), (["foo"], "bar"), (["foo", "bar"], "baz")]
)
def test_set_validator_fails(field_set, field):
    validator = SetValidator(field_set)
    with pytest.raises(ValidationException):
        validator.validate(field)

    assert validator.bad == {field}


@pytest.mark.parametrize(
    "fields, row, unique_with",
    [
        ([], {}, []),
        (["foo", "bar"], {}, []),
        (["foo", ""], {}, []),
        (["foo", "foo"], FakeRow({"some_field": ["bar", "baz"]}), ["some_field"]),
        (
            ["foo", "foo", "foo"],
            FakeRow(
                {
                    "some_field": ["bar", "baz", "bar"],
                    "some_other_field": ["baz", "bar", "bar"],
                }
            ),
            ["some_field", "some_other_field"],
        ),
    ],
)
def test_unique_validator_works(fields, row, unique_with):
    validator = UniqueValidator(unique_with=unique_with)
    for field in fields:
        validator.validate(field, row)


@pytest.mark.parametrize("fields, row, unique_with", [(["foo", "bar", "", ""], {}, [])])
def test_unique_validator_supports_empty_ok(fields, row, unique_with):
    validator = UniqueValidator(unique_with=unique_with, empty_ok=True)
    for field in fields:
        validator.validate(field, row)


@pytest.mark.parametrize(
    "fields, row, unique_with, exception, bad",
    [
        (["foo", "bar", "bar"], {}, [], ValidationException, {("bar",)}),
        (["foo", "", ""], {}, [], ValidationException, {("",)}),
        (["", "", ""], {}, [], ValidationException, {("",)}),
        (
            ["foo", "foo"],
            FakeRow({"some_field": ["bar", "bar"]}),
            ["some_field"],
            ValidationException,
            {("foo", "bar")},
        ),
        (
            ["foo", "foo"],
            FakeRow({"some_field": ["bar", "bar"]}),
            ["other_field"],
            BadValidatorException,
            set(),
        ),
    ],
)
def test_unique_validator_fails(fields, row, unique_with, exception, bad):
    validator = UniqueValidator(unique_with=unique_with)
    with pytest.raises(exception):
        for field in fields:
            validator.validate(field, row)

    assert validator.bad == bad


@pytest.mark.parametrize("pattern, field", [(r"foo.*", "foo"), (r"foo.*", "foobar")])
def test_regex_validator_works(pattern, field):
    RegexValidator(pattern).validate(field)


@pytest.mark.parametrize("pattern, field", [(r"[^\s]+", "foo bar")])
def test_regex_validator_full_works(pattern, field):
    validator = RegexValidator(pattern, full=True)
    with pytest.raises(ValidationException):
        validator.validate(field)

    assert validator.bad == {field}


@pytest.mark.parametrize("pattern, field", [(r"foo.*", "afoo"), (r"^$", "foo")])
def test_regex_validator_fails(pattern, field):
    validator = RegexValidator(pattern)
    with pytest.raises(ValidationException):
        validator.validate(field)

    assert validator.bad == {field}


def test_range_validator_works():
    RangeValidator(0, 100).validate("42")


def test_range_validator_fails():
    validator = RangeValidator(0, 100)
    with pytest.raises(ValidationException):
        validator.validate("-42")

    assert validator.bad == {"-42"}


def test_range_validator_handles_bad_values():
    validator = RangeValidator(0, 100)
    with pytest.raises(ValidationException):
        validator.validate("foobar")

    assert validator.bad == {"foobar"}


def test_empty_validator_works():
    EmptyValidator().validate("")


def test_empty_validator_fails():
    validator = EmptyValidator()
    with pytest.raises(ValidationException):
        validator.validate("foo")

    assert validator.bad == {"foo"}


def test_non_empty_validator_works():
    NotEmptyValidator().validate("foo")


def test_non_empty_validator_fails():
    validator = NotEmptyValidator()
    with pytest.raises(ValidationException):
        validator.validate("")

    assert validator.bad is True


def test_ignore_validator():
    validator = Ignore()
    validator.validate("foo")
    assert validator.bad is None


def test_base_class_raises():
    validator = Validator()

    with pytest.raises(NotImplementedError):
        validator.bad()

    with pytest.raises(NotImplementedError):
        validator.validate(stub(), stub())


def test_row_length_validator_works():
    validator = RowLengthValidator()
    validator.validate({"Field One": "1", "Field Two": "Two"})


@pytest.mark.parametrize(
    "row",
    [
        {"Field One": "1", "Field Two": None},
        {"Field One": "1", "Field Two": "2", None: ["3", "4"]},
    ],
)
def test_row_length_validator_fails(row):
    validator = RowLengthValidator()
    with pytest.raises(ValidationException):
        validator.validate(row)

    assert validator.bad == [row]


def test_base_row_validator_raises():
    validator = RowValidator()

    with pytest.raises(NotImplementedError):
        validator.bad

    with pytest.raises(NotImplementedError):
        validator.validate(stub())


@pytest.mark.parametrize(
    "validator_class,args",
    [
        (SetValidator, [["foo"]]),
        (RegexValidator, [r"foo.*"]),
        (IntValidator, []),
        (FloatValidator, []),
        (RangeValidator, [0, 42]),
        (UniqueValidator, []),
    ],
)
def test_all_validators_support_empty_ok(validator_class, args):
    validator = validator_class(*args, empty_ok=True)
    validator.validate("")


@pytest.mark.parametrize(
    "a_set, max_len, stringified",
    [
        ({"A", "B", "C"}, 4, "{'A', 'B', 'C'}"),
        ({"A", "B", "C"}, 3, "{'A', 'B', 'C'}"),
        ({"A", "B", "C"}, 2, "{'A', 'B'} (1 more suppressed)"),
        ({"A", "B", "C"}, 0, "{} (3 more suppressed)"),
        ({}, 5, "{}"),
    ],
)
def test_stringify_set(a_set, max_len, stringified):
    assert _stringify_set(a_set, max_len) == stringified



# encoding: utf-8

import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

__version__ = "0.0.25"


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ["-x", "tests"]

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


def readme():
    with open("README.rst") as f:
        return f.read()


setup(
    name="vladiate",
    version=__version__,
    description="Vladiate is a strict validation tool for CSV files",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Intended Audience :: Developers",
    ],
    keywords="validate CSV vampires",
    author="Dustin Ingram",
    author_email="github@dustingram.com",
    url="http://github.com/di/vladiate",
    license="MIT",
    long_description=readme(),
    packages=find_packages(exclude=["examples", "tests"]),
    include_package_data=True,
    zip_safe=False,
    install_requires=[],
    extras_require={"s3": ["boto"]},
    tests_require=["pretend", "pytest", 'black;python_version>="3.6"'],
    cmdclass={"test": PyTest},
    entry_points={
        "console_scripts": [
            "vladiate = vladiate.main:main",
        ]
    },
)



from multiprocessing import Pool, Queue
from vladiate import Vlad
from vladiate import logs
from vladiate import exits

import os
import sys
import inspect
from argparse import ArgumentParser
from pkg_resources import get_distribution


def parse_args():
    """
    Handle command-line arguments with argparse.ArgumentParser
    Return list of arguments, largely for use in `parse_arguments`.
    """

    # Initialize
    parser = ArgumentParser(
        description="vladiate [options] [VladClass [VladClass2 ... ]]"
    )

    parser.add_argument(
        "vlads",
        metavar="vlads",
        type=str,
        nargs="*",
        help="A list of Vlad classes to validate",
    )

    # Specify the vladfile to be something other than vladfile.py
    parser.add_argument(
        "-f",
        "--vladfile",
        dest="vladfile",
        default="vladfile",
        help="Python module to import, e.g. '../other.py'. Default: vladfile",
    )

    # List vladiate commands found in loaded vladiate files/source files
    parser.add_argument(
        "-l",
        "--list",
        action="store_true",
        dest="list_commands",
        default=False,
        help="Show list of possible vladiate classes and exit",
    )

    # Version number
    parser.add_argument(
        "-V",
        "--version",
        action="store_true",
        dest="show_version",
        default=False,
        help="show program's version number and exit",
    )

    # Maximum number of processes to attempt to use
    parser.add_argument(
        "-p",
        "--processes",
        dest="processes",
        default=1,
        type=int,
        help="attempt to use this number of processes",
    )

    # Disable vladiate classes console log output
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        dest="quiet",
        default=False,
        help="disable console log output generated by validations",
    )

    return parser.parse_args()


def is_vlad(tup):
    """
    Takes (name, object) tuple, returns True if it's a public Vlad subclass.
    """
    name, item = tup
    return bool(
        inspect.isclass(item)
        and issubclass(item, Vlad)
        and hasattr(item, "source")
        and getattr(item, "source")
        and hasattr(item, "validators")
        and not name.startswith("_")
    )


def _is_package(path):
    """
    Is the given path a Python package?
    """
    return os.path.isdir(path) and os.path.exists(os.path.join(path, "__init__.py"))


def find_vladfile(vladfile, path="."):
    """
    Attempt to locate a vladfile, either explicitly or by searching parent dirs.
    """
    assert os.path.isdir(path)
    # Obtain env value
    names = [vladfile]
    # Create .py version if necessary
    if not names[0].endswith(".py"):
        names += [names[0] + ".py"]
    # Does the name contain path elements?
    if os.path.dirname(names[0]):
        # If so, expand home-directory markers and test for existence
        for name in names:
            expanded = os.path.expanduser(name)
            if os.path.exists(expanded):
                if name.endswith(".py") or _is_package(expanded):
                    return os.path.abspath(expanded)
    else:
        for name in names:
            joined = os.path.join(path, name)
            if os.path.exists(joined):
                if name.endswith(".py") or _is_package(joined):
                    return os.path.abspath(joined)
    # Implicit 'return None' if nothing was found


def load_vladfile(path):
    """
    Import given vladfile path and return (docstring, callables).
    Specifically, the vladfile's ``__doc__`` attribute (a string) and a
    dictionary of ``{'name': callable}`` containing all callables which pass
    the "is a vlad" test.
    """
    # Get directory and vladfile name
    directory, vladfile = os.path.split(path)
    # If the directory isn't in the PYTHONPATH, add it so our import will work
    added_to_path = False
    index = None
    if directory not in sys.path:
        sys.path.insert(0, directory)
        added_to_path = True
    # If the directory IS in the PYTHONPATH, move it to the front temporarily,
    # otherwise other vladfiles -- like vlads's own -- may scoop the intended
    # one.
    else:
        i = sys.path.index(directory)
        if i != 0:
            # Store index for later restoration
            index = i
            # Add to front, then remove from original position
            sys.path.insert(0, directory)
            del sys.path[i + 1]
    # Perform the import (trimming off the .py)
    imported = __import__(os.path.splitext(vladfile)[0])
    # Remove directory from path if we added it ourselves (just to be neat)
    if added_to_path:
        del sys.path[0]
    # Put back in original index if we moved it
    if index is not None:
        sys.path.insert(index + 1, directory)
        del sys.path[0]
    # Return our two-tuple
    vlads = dict(filter(is_vlad, vars(imported).items()))
    return imported.__doc__, vlads


def _vladiate(vlad):
    global result_queue
    result_queue.put(
        vlad(
            vlad.source, validators=vlad.validators, quiet=parse_args().quiet
        ).validate()
    )


result_queue = Queue()


def main():
    arguments = parse_args()
    logger = logs.logger

    if arguments.show_version:
        print("Vladiate %s" % (get_distribution("vladiate").version,))
        return exits.OK

    vladfile = find_vladfile(arguments.vladfile)
    if not vladfile:
        logger.error(
            "Could not find any vladfile! Ensure file ends in '.py' and see "
            "--help for available options."
        )
        return exits.NOINPUT

    docstring, vlads = load_vladfile(vladfile)

    if arguments.list_commands:
        logger.info("Available vlads:")
        for name in vlads:
            logger.info("    " + name)
        return exits.OK

    if not vlads:
        logger.error("No vlad class found!")
        return exits.NOINPUT

    # make sure specified vlad exists
    if arguments.vlads:
        missing = set(arguments.vlads) - set(vlads.keys())
        if missing:
            logger.error("Unknown vlad(s): %s\n" % (", ".join(missing)))
            return exits.UNAVAILABLE
        else:
            names = set(arguments.vlads) & set(vlads.keys())
            vlad_classes = [vlads[n] for n in names]
    else:
        vlad_classes = vlads.values()

    # validate all the vlads, and collect the validations for a good exit
    # return code
    all_passed = True
    if arguments.processes == 1:
        for vlad in vlad_classes:
            passed = vlad(source=vlad.source, quiet=arguments.quiet).validate()
            all_passed = all_passed and passed

    else:
        proc_pool = Pool(
            arguments.processes
            if arguments.processes <= len(vlad_classes)
            else len(vlad_classes)
        )
        proc_pool.map(_vladiate, vlad_classes)
        while not result_queue.empty():
            passed = result_queue.get()
            all_passed = all_passed and passed

    return exits.OK if all_passed else exits.DATAERR


def run(name):
    if name == "__main__":
        exit(main())


run(__name__)
